import os

from pathlib import Path
from typing import Any, Dict

import chevron


dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
templates_path = dir_path / "templates"


def generate_pyproject(swagger: Dict[str, Any], out_file: Path, project_name: str) -> None:
    """
    Generate `pyproject.toml` file.
    """
    version = swagger["info"]["version"]

    with open(templates_path / "pyproject.toml.mustache", "r") as f:
        toml_str = chevron.render(f, {"version": version, "project_name": project_name})

    with open(out_file, "w+") as f:
        f.write(toml_str)



import argparse
import json
import os
import shutil

from pathlib import Path

from python_client_generator.utils import dereference_swagger

from .generate_apis import generate_apis
from .generate_base_client import generate_base_client
from .generate_models import generate_models
from .generate_pyproject import generate_pyproject


dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
templates_path = dir_path / "templates"


def main() -> None:
    parser = argparse.ArgumentParser(description="Generates an httpx-based Python client.")
    parser.add_argument("--open-api", type=str)
    parser.add_argument("--package-name", type=str)
    parser.add_argument("--project-name", type=str)
    parser.add_argument("--outdir", type=str, default="clients/")
    parser.add_argument("--group-by-tags", action="store_true")
    parser.add_argument("--sync", action="store_true")

    args = parser.parse_args()

    with open(args.open_api, "r") as f:
        swagger = json.load(f)
    dereferenced_swagger = dereference_swagger(swagger, swagger)

    # Create root directory
    path = Path(args.outdir)
    path.mkdir(parents=True, exist_ok=True)

    generate_pyproject(dereferenced_swagger, path / "pyproject.toml", args.project_name)

    # Create package directory
    package_path = path / Path(args.package_name)
    package_path.mkdir(parents=True, exist_ok=True)

    # Generate package files
    generate_models(dereferenced_swagger, package_path / "models.py")
    generate_base_client(package_path / "base_client.py", sync=args.sync)
    generate_apis(dereferenced_swagger, package_path / "apis.py", args.group_by_tags, args.sync)
    shutil.copyfile(templates_path / "__init__.py", package_path / "__init__.py")



import os
import re

from pathlib import Path
from typing import Any, Dict, List

import chevron

from .utils import resolve_type, sanitize_name, serialize_to_python_code


dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
templates_path = dir_path / "templates"


def resolve_field_args(property: Dict[str, Any]) -> Dict[str, Any]:
    """
    Map OpenAPI properties to Pydantic Field function args
    """
    args = {}
    if "maxLength" in property:
        args["max_length"] = property["maxLength"]
    if "minLength" in property:
        args["min_length"] = property["minLength"]
    # TODO: decide later if we want the description field or a comment above the line
    # if "description" in property:
    #     args_dict["description"] = property["description"]
    if "default" in property:
        args["default"] = property["default"]

    return args


def serialize_args_dict(args_dict: Dict[str, Any]) -> str:
    args_list = [f"{k}={serialize_to_python_code(v)}" for k, v in args_dict.items()]
    return ", ".join(args_list)


def _get_schema_references(schema: Dict[str, Any]) -> List[str]:
    union_keys = list(set(["allOf", "anyOf", "oneOf"]) & set(schema.keys()))
    if union_keys:
        arr = []
        for p_sub_schema in schema[union_keys[0]]:
            arr += _get_schema_references(p_sub_schema)
        return arr
    elif "type" not in schema:
        return []
    elif schema["type"] == "array":
        return _get_schema_references(schema["items"])
    elif schema["type"] == "object" or "enum" in schema:
        return [schema["title"]]
    else:
        return []


def get_references(model: Dict[str, Any]) -> List[str]:
    """
    Get a list of dependencies for a model

    These come from either their properties or unionized referencing.
    """
    refs = []
    union_keys = list(set(["allOf", "anyOf", "oneOf"]) & set(model.keys()))
    if union_keys:
        return _get_schema_references(model)
    else:
        # Must have properties
        for p_schema in model["properties"].values():
            refs += _get_schema_references(p_schema)

    return [sanitize_name(r) for r in refs]


def get_fields(schema: Dict[str, Any]) -> List[Dict[str, Any]]:
    union_keys = list(set(["allOf", "anyOf", "oneOf"]) & set(schema.keys()))
    if union_keys:
        # Handle union cases by creating a __root__ defined model
        return [{"name": "__root__", "type": resolve_type(schema)}]
    else:
        return [
            {
                "name": k,
                "type": resolve_type(v, use_literals=True),
                "optional": "required" not in schema or k not in schema["required"],
                "field_args": serialize_args_dict(resolve_field_args(v)),
            }
            for k, v in schema["properties"].items()
        ]


def _strip_nonexistant_refs(objects: List[Dict[str, Any]]) -> None:
    """
    Remove any references to names not in the objects list
    """
    names = [o["name"] for o in objects]
    for o in objects:
        o["refs"] = [ref for ref in o["refs"] if ref in names]


def _get_ref_index(objects: List[Dict[str, Any]], name: str) -> int:
    return next(i for i, o in enumerate(objects) if o["name"] == name)


def _object_has_binary_properties(object: Dict[str, Any]) -> bool:
    """
    Determine if an object has some properties which are binary
    """
    for p in object.get("properties", {}).values():
        if p.get("type") == "string" and p.get("format") == "binary":
            return True
    return False


def _sort_models(objects: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Resolve object dependencies by walking through the objects list
    and moving any dependencies which are further along in the list
    ahead of us.
    """
    i = 0
    while i < len(objects):
        sorted = False
        for ref in objects[i]["refs"]:
            ref_index = _get_ref_index(objects, ref)
            if ref_index > i:
                objects.insert(i, objects.pop(ref_index))
                sorted = True

        if not sorted:
            i += 1  # Ensure shifted references reprocessed

    return objects


def get_models(schemas: Dict[str, Any]) -> List[Dict[str, Any]]:
    objects = (v for v in schemas.values() if "type" not in v or v["type"] == "object")

    models = []
    for o in objects:
        # Skip models with "binary" properties as these are related to file uploads
        # and we handle them with function arguments on the API
        if _object_has_binary_properties(o):
            continue

        p: Dict[str, Any] = {}
        p["refs"] = get_references(o)
        p["name"] = sanitize_name(o["title"])
        p["fields"] = get_fields(o)
        models.append(p)

    _strip_nonexistant_refs(models)
    _sort_models(models)
    return models


def _enum_val_to_name(value: Any) -> str:
    """
    Generate name for each enumeration value.
    """
    if isinstance(value, (int, float)):
        return "NUMBER_" + str(float(value)).replace(".", "_DOT_")
    elif isinstance(value, (str)):
        # TODO: use proper regex replace to insure we don't create variable names non
        # compliant with Python syntax
        char_list = [" ", "-", ".", ",", "&"]
        for c in char_list:
            value = value.replace(c, "_")
        return "".join(re.findall(r"[A-Za-z0-9]+[A-Za-z0-9_]*", value.upper()))
    else:
        raise Exception(f"Unsupported enum value type: {value.__class__}")


def get_enums(schemas: Dict[str, Any]) -> List[Dict[str, Any]]:
    objects = {k: v for k, v in schemas.items() if "enum" in v}

    enums = []
    for _, o in objects.items():
        p: Dict[str, Any] = {}
        p["name"] = sanitize_name(o["title"])
        p["type"] = resolve_type(o)
        p["fields"] = [
            {
                "name": _enum_val_to_name(v),
                "value": serialize_to_python_code(v),
            }
            for v in o["enum"]
        ]
        enums.append(p)

    return enums


def generate_models(swagger: Dict[str, Any], out_file: Path) -> None:
    """
    Generate enums and Pydantic models from the dereferenced OpenAPI file.
    """
    schemas = swagger["components"]["schemas"]

    models = get_models(schemas)
    enums = get_enums(schemas)

    with open(templates_path / "models.py.mustache", "r") as f:
        models_str = chevron.render(f, {"enums": enums, "models": models})

    with open(out_file, "w+") as f:
        f.write(models_str)



import os

from pathlib import Path

import chevron


dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
templates_path = dir_path / "templates"


def generate_base_client(out_file: Path, sync: bool) -> None:
    """
    Generate the root API client to be used by all other functions.
    """
    with open(templates_path / "base_client.py.mustache", "r") as f:
        toml_str = chevron.render(f, {"async": not sync})

    with open(out_file, "w+") as f:
        f.write(toml_str)



import re

from typing import Any, Dict, List


def lookup_by_ref_parts(obj: Dict[str, Any], ref_parts: List[str]) -> Dict[str, Any]:
    child = obj[ref_parts[0]]
    if len(ref_parts) == 1:
        return child
    else:
        return lookup_by_ref_parts(child, ref_parts[1:])


def dereference(schema: Dict[str, Any], ref: str) -> Dict[str, Any]:
    ref_parts = ref.split("/")[1:]
    return lookup_by_ref_parts(schema, ref_parts)


def dereference_swagger(current: Any, original: Dict[str, Any]) -> Any:
    """
    Dereference an OpenAPI file in place.
    Note: We only dereference the references that are values of a dict, not a list
    """
    if isinstance(current, dict):
        if "$ref" in current:
            current = dereference(original, current["$ref"])

        return {k: dereference_swagger(v, original) for k, v in current.items()}

    elif isinstance(current, list):
        return [dereference_swagger(v, original) for v in current]
    else:
        return current


def serialize_to_python_code(obj: Any) -> str:
    # TODO: Support List/Dict nesting
    if isinstance(obj, (int, float, bool, list, dict)):
        return str(obj)
    else:
        return f'"{obj}"'


def sanitize_name(name: str) -> str:
    return "".join(re.findall(r"[A-Za-z0-9]+[A-Za-z0-9_]*", name))


def resolve_type(schema: Dict[str, Any], depth: int = 0, use_literals: bool = False) -> str:
    """
    Resolve Python type for a given schema
    """

    if "title" in schema and depth > 0:
        # Just return title for any nested types to prevent elaborating
        # Unions, etc.
        return schema["title"]

    union_keys = list(set(["allOf", "anyOf", "oneOf"]) & set(schema.keys()))
    if union_keys:
        # Handle union cases
        result: List[str] = []
        for sub_schema in schema[union_keys[0]]:
            type_ = resolve_type(sub_schema, depth + 1)
            result.append(type_)
        if len(result) > 1:
            return f"Union[{', '.join(result)}]"
        elif len(result) == 1:
            return result[0]
        else:
            raise Exception(f"Empty union type detected for '{schema['title']}'")
    elif "type" not in schema:
        return "Any"
    elif schema["type"] == "object":
        if "properties" in schema:
            return sanitize_name(schema["title"])
        else:
            return "Dict[str, Any]"
    elif schema["type"] == "string":
        if schema.get("format") == "date-time":
            return "datetime"
        elif schema.get("format") == "uuid":
            return "UUID"
        elif schema.get("format") == "binary":
            return "httpx._types.FileTypes"
        elif use_literals and len(schema.get("enum", [])) == 1:
            return f"Literal['{schema['enum'][0]}']"
        else:
            return "str"
    elif schema["type"] == "boolean":
        return "bool"
    elif schema["type"] == "integer":
        return "int"
    elif schema["type"] == "number":
        return "float"
    elif schema["type"] == "array":
        return "List[" + resolve_type(schema["items"], depth + 1) + "]"

    raise Exception("property: ", schema)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_client_generator', 'python_client_generator.templates']

package_data = \
{'': ['*']}

install_requires = \
['chevron>=0.14.0,<0.15.0', 'pydantic>=1.9.1,<2.0.0']

entry_points = \
{'console_scripts': ['poetry = poetry.console:main']}

setup_kwargs = {
    'name': 'python-client-generator',
    'version': '1.0.6',
    'description': 'Python package to generate an httpx-based client off an OpenAPI spec',
    'long_description': '# python-client-generator\n\nPython package to generate an [httpx](https://github.com/encode/httpx)- and\n[pydantic](https://github.com/pydantic/pydantic)-based async (or sync) \nclient off an OpenAPI spec.\n\n```mermaid\nflowchart LR\n    generator["python-client-generator"]\n    app["REST API app"]\n    package["app HTTP client"]\n\n    app -- "OpenAPI json" --> generator\n    generator -- "generates" --> package\n```\n\n\n## Using the generator\n\n```bash\npython -m python_client_generator --open-api openapi.json --package-name foo_bar --project-name foo-bar --outdir clients\n```\n\nThis will produce a Python package with the following structure:\n```bash\nclients\n├── foo_bar\n│\xa0\xa0 ├── __init__.py\n│\xa0\xa0 ├── apis.py\n│\xa0\xa0 ├── base_client.py\n│\xa0\xa0 └── models.py\n└── pyproject.toml\n```\n\n### Using PATCH functions from the generator\n\nWhen calling one of the generated update functions that uses an HTTP `PATCH` method, you\'ll\nprobably want to pass the additional argument `body_serializer_args={"exclude_unset": True}`. This\nwill ensure that only the fields that are set in the update object are sent to the API. Example:\n\n```python\nawait api_client.update_contact_v1_contacts__contact_id__patch(\n                body=patch_body,\n                contact_id=contact.id,\n                tenant=tenant,\n                body_serializer_args={"exclude_unset": True}\n)\n```\n\n\n## Contributing\nInstall the dependencies with `poetry`:\n```shell\npoetry install\n```\n\nFormat the code:\n```shell\nmake format\n```\n\nTest the code locally:\n```shell\npoetry run pytest\n```\n\n\n### Commiting\n\nWe use commitlint in the CI/CD to make sure all commit messages adhere to [conventionalcommits](https://www.conventionalcommits.org/en/v1.0.0/). See `.commitlintrc.js`, `.releaserc` and `.czrc.json` for specific implementation details.\n\nAs per the default commitlint settings for conventionalcommits ([see here](https://github.com/conventional-changelog/commitlint))\nthe following commit types may be used:\n\n  - `build`\n  - `chore`\n  - `ci`\n  - `docs`\n  - `feat`\n  - `fix`\n  - `perf`\n  - `refactor`\n  - `revert`\n  - `style`\n  - `test`\n\nOf which the following will cause a release (one of these types *must* be used if you are submitting code\nthat you expect to be deployed after merging):\n\n  - `build`\n  - `ci`\n  - `docs(api)`\n  - `feat`\n  - `fix`\n  - `perf`\n  - `refactor`\n  - `revert`\n  \nTo ensure that your commits always conform to the above format, you can install `commitizen`:\n```shell\nnpm i -g commitizen\n```\n',
    'author': 'Bogdan Petre',
    'author_email': 'bogdan.petre@sennder.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import chevron

from .utils import resolve_type, sanitize_name, serialize_to_python_code


dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
templates_path = dir_path / "templates"


def resolve_property_type(property: Dict[str, Any]) -> str:
    type_ = resolve_type(property["schema"])
    return (
        f"Optional[{type_}]"
        if property.get("required") is False and property["schema"]["type"] != "boolean"
        else type_
    )


def resolve_property_default(property: Dict[str, Any]) -> Optional[str]:
    if property.get("required") is not False:
        return None

    if "default" in property["schema"]:
        return serialize_to_python_code(property["schema"]["default"])

    return "None"


def get_return_type(responses: Dict[str, Any]) -> Optional[str]:
    successful_responses = [v for k, v in responses.items() if int(k) >= 200 and int(k) < 300]
    if len(successful_responses) != 1:
        raise Exception("Incorrect number of successful responses:", len(successful_responses))

    schema = successful_responses[0]["content"]["application/json"]["schema"]
    if not schema:
        return None

    return sanitize_name(schema["title"])


def _get_request_body_params(method: Dict[str, Any]) -> List[Dict[str, Any]]:
    args = []

    content = method["requestBody"]["content"]
    if "application/json" in content:
        schema = content["application/json"]["schema"]
        args.append(
            {
                "name": "body",  # TODO: think about naming
                "schema": schema,
            }
        )
    else:
        # Create argument for each multipart upload property
        schema = content["multipart/form-data"]["schema"]
        for k, v in schema["properties"].items():
            args.append({"name": k, "schema": v})

    return args


def get_function_args(method: Dict[str, Any]) -> List[Dict[str, Any]]:
    params = []

    # Get request body argument(s)
    if "requestBody" in method:
        params += _get_request_body_params(method)

    # Get path, query, and header parameters - sorting first by required and then non-required
    keys = ["path", "query", "header"]
    parameters = method.get("parameters", [])
    for k in keys:
        params += [p for p in parameters if p["in"] == k and p["required"]]
    for k in keys:
        params += [p for p in parameters if p["in"] == k and not p["required"]]

    # Convert params to args format required for templating
    return [
        {
            "name": p["name"],
            "type": resolve_property_type(p),
            "default": resolve_property_default(p),
            "has_default": resolve_property_default(p) is not None,
        }
        for p in params
    ]


def get_params_by_type(method: Dict[str, Any], type_: str) -> List[Dict[str, str]]:
    """
    :param str type_: `query` or `header`
    """
    params = method.get("parameters", [])
    return [{"name": p["name"]} for p in params if p["in"] == type_]


def has_json_body(method: Dict[str, Any]) -> bool:
    try:
        _ = method["requestBody"]["content"]["application/json"]
        return True
    except KeyError:
        return False


def get_multipart_properties(
    method: Dict[str, Any]
) -> Tuple[List[Dict[str, str]], List[Dict[str, str]]]:
    """
    Parse method for any multipart form upload file and data properties.
    """
    try:
        schema = method["requestBody"]["content"]["multipart/form-data"]["schema"]
    except KeyError:
        return ([], [])

    files, data = [], []
    for k, v in schema["properties"].items():
        if v.get("type") == "string" and v.get("format") == "binary":
            files.append({"name": k})
        else:
            data.append({"name": k})

    return (files, data)


@dataclass
class TaggedEndpointDefinition:
    path_name: str
    method_name: str
    method: Dict[str, Any]


def get_endpoints(
    endpoint_defs: Iterable[TaggedEndpointDefinition], sync: bool
) -> List[Dict[str, Any]]:
    endpoints: List[Dict[str, Any]] = []

    for e_def in endpoint_defs:
        e: Dict[str, Any] = {}
        files, data = get_multipart_properties(e_def.method)
        e["name"] = e_def.method["operationId"]
        e["method"] = e_def.method_name.upper()
        e["async"] = not sync
        e["args"] = get_function_args(e_def.method)
        e["path"] = f'f"{e_def.path_name}"' if "{" in e_def.path_name else f'"{e_def.path_name}"'
        e["return_type"] = get_return_type(e_def.method["responses"])
        e["docs"] = e_def.method.get("description", "").replace("\n", "\n        ")
        e["query_params"] = get_params_by_type(e_def.method, "query")
        e["has_query_params"] = len(e["query_params"]) > 0
        e["header_params"] = get_params_by_type(e_def.method, "header")
        e["has_header_params"] = len(e["header_params"]) > 0
        e["has_json_body"] = has_json_body(e_def.method)
        e["multipart_data"] = data
        e["has_multipart_data"] = len(data) > 0
        e["multipart_files"] = files
        e["has_multipart_files"] = len(files) > 0

        endpoints.append(e)

    return endpoints


def get_apis(swagger: Dict[str, Any], group_by_tags: bool, sync: bool) -> List[Dict[str, Any]]:
    paths = swagger["paths"]

    # Group endpoints by tags
    api_groups: Dict[Optional[str], List[TaggedEndpointDefinition]] = {}
    for path_name, path in paths.items():
        for method_name, method in path.items():
            tags = method.get("tags", []) if group_by_tags else [None]
            for tag in tags:
                if tag not in api_groups:
                    api_groups[tag] = []
                api_groups[tag].append(TaggedEndpointDefinition(path_name, method_name, method))

    # Create groups of "apis":
    return [
        {
            "class_name": f"{tag.capitalize() if tag else ''}Api",
            "tag": tag,
            "async": not sync,
            "endpoints": get_endpoints(endpoint_defs, sync),
        }
        for tag, endpoint_defs in api_groups.items()
    ]


def generate_apis(swagger: Dict[str, Any], out_file: Path, group_by_tags: bool, sync: bool) -> None:
    """
    Generate functions for each path in the dereferenced OpenAPI input file.
    """
    apis = get_apis(swagger, group_by_tags, sync)

    with open(templates_path / "apis.py.mustache", "r") as f:
        models_str = chevron.render(f, {"apis": apis})

    with open(out_file, "w+") as f:
        f.write(models_str)
